/****************************/
/*   	BG3D EXPORT.C 		*/
/* (c)1999 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/****************************/
/*    EXTERNALS             */
/****************************/

#include <QD3D.h>
#include <QD3DErrors.h>
#include <QD3DGeometry.h>
#include <QD3DGroup.h>
#include "QD3DMath.h"
#include "QD3DShader.h"
#include "QD3DStyle.h"
#include <QD3DTransform.h>

#include "myglobals.h"
#include "objects.h"
#include "mymenus.h"
#include "mywindows.h"
#include "myevents.h"
#include "misc.h"
#include "bg3d.h"
#include "qd3d_support.h"
#include "3dmf.h"
#include "dialogStuff.h"
#include "optimize.h"


extern	ObjNode					*gModelObjNode;
extern	Boolean					gCurrentIsOptimized;
extern	short					gNumMaterialTypes;
extern	MaterialType			gMaterialTypeList[];
extern	TQ3Object			gModelGroup;

/****************************/
/*    PROTOTYPES            */
/****************************/

static Boolean GetBG3DOutputFileName(FSSpec *myFSSpec);
static void WriteBG3DHeader(short refNum);
static void WriteBG3DMaterials(short refNum);
static void WriteAMaterial(short refNum, MaterialType *mat);
static void WriteTexture(short refNum, MaterialType *srcMat);
static void WriteBG3DGeometry(short refNum);
static void WriteATriMesh(short refNum, int i, TQ3Object triMeshObj);
static void WriteNormalArray(short refNum, TQ3Vector3D *vectors, int numPoints);
static void WriteUVArray(short refNum, TQ3Param2D *uvs, int numPoints);
static void WriteColorArray(short refNum, TQ3ColorRGB *colors, int numPoints);
static void WriteBG3DEndTag(short refNum);


/****************************/
/*    CONSTANTS             */
/****************************/


/*********************/
/*    VARIABLES      */
/*********************/

static FSSpec	gLastBG3D_FSSpec;


/********************** EXPORT BG3D **************************/

void ExportBG3D(void)
{
FSSpec	myFSSpec;
OSErr	iErr;
short	refNum;


	if (!gCurrentIsOptimized)
	{
		DoAlert("\pMust optimize first!");
		return;
	}

			/* LET USER CREATE NEW FILE */
			
	if (!GetBG3DOutputFileName(&myFSSpec))
		return;
		
			/* OPEN THE FILE FOR WRITING */
			
	iErr = FSpOpenDF(&myFSSpec, fsWrPerm, &refNum);
	if (iErr)
		DoFatalAlert("\pExportBG3D: FSpOpenDF failed!");
		
	
			/* WRITE ALL THE COMPONENTS */
			
	WriteBG3DHeader(refNum);
	WriteBG3DMaterials(refNum);
	WriteBG3DGeometry(refNum);
	
			/* CLOSE FILE */
			
	WriteBG3DEndTag(refNum);
	FSClose(refNum);
}


/******************* GET BG3D OUTPUT FILE NAME ********************/
//
// OUTPUT: false = error occurred
//			FSSpec = file info
//

static Boolean GetBG3DOutputFileName(FSSpec *myFSSpec)
{
StandardFileReply	reply;
	
		
	StandardPutFile("\pExport BG3D file",gLastBG3D_FSSpec.name,&reply);
	if (!reply.sfGood)											// see if cancelled 
		return(false);

	*myFSSpec = reply.sfFile;									// copy FSSpec back
	
	if (reply.sfReplacing)
		FSpDelete(myFSSpec);									// delete any old one
		
	FSpCreate(myFSSpec, 'OP40', 'BG3D', smSystemScript);		// create new file
	
	return(true);
}


#pragma mark -

/********************** WRITE BG3D HEADER **************************/

static void WriteBG3DHeader(short refNum)
{
BG3DHeaderType	headerData;
char			s[16] = "BG3D 1.0        ";
long			count;


			/* SET HEADER FIELDS */
			
	BlockMove(&s[0], &headerData.headerString[0], 16);		// set header string
	
	headerData.version.majorRev 		= 0x01;
	headerData.version.minorAndBugRev 	= 0x00;
	headerData.version.stage 			= 0x00;
	headerData.version.nonRelRev	 	= 0x00;


			/* WRITE HEADER TO FILE */
	
	count = sizeof(BG3DHeaderType);			
	if (FSWrite(refNum, &count, &headerData) != noErr)
		DoFatalAlert("\pWriteBG3DHeader: FSWrite failed!");
}

/********************** WRITE BG3D END TAG **************************/

static void WriteBG3DEndTag(short refNum)
{
u_long			tag;
long			count;

	tag = BG3D_TAGTYPE_ENDFILE;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);							// write tag
}
	
	
/********************** WRITE BG3D MATERIALS **************************/

static void WriteBG3DMaterials(short refNum)
{
long			i;

	if (gNumMaterialTypes == 0)										// see if there are any materials
		return;
		
		/***************************/
		/* WRITE EACH MATERIAL OUT */
		/***************************/
			
	for (i = 0; i < gNumMaterialTypes; i++)
	{
		WriteAMaterial(refNum, &gMaterialTypeList[i]);
	}
}	



/********************* WRITE A MATERIAL ***********************/

static void WriteAMaterial(short refNum, MaterialType *srcMat)
{
long			count;
u_long			tag,flags;
float			alpha;
GLfloat			color[4];

		/*************************/
		/* WRITE MATERIALS FLAGS */
		/*************************/

			/* SET FLAGS */
			
	flags = 0;
	
	if (srcMat->isTextured)
		flags |= BG3D_MATERIALFLAG_TEXTURED;

	if (srcMat->isTransparent)
		flags |= BG3D_MATERIALFLAG_TRANSPARENT;


		/* WRITE TAG & FLAGS TO FILE */
	
	tag = BG3D_TAGTYPE_MATERIALFLAGS;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);							// write tag
	count = sizeof(flags);
	FSWrite(refNum, &count, &flags);						// write flags


		/****************************/
		/* WRITE DIFFUSE COLOR INFO */
		/****************************/
		
		/* CALCULATE ALPHA */
		
	if (srcMat->isTransparent)
		alpha = srcMat->transColor.r;			// use src transparency value
	else
		alpha = 1;								// otherwise, assume opaque

		/* SET COLOR */
				
	if (srcMat->isTextured)
	{
		color[0] = 1;							// set diffuse color to white
		color[1] = 1;
		color[2] = 1;
		color[3] = alpha;
	}
	else
	{
		color[0] = srcMat->color.r;				
		color[1] = srcMat->color.g;
		color[2] = srcMat->color.b;				
		color[3] = alpha;
	}			
		
		/* WRITE TAG & COLOR TO FILE */
	
	tag = BG3D_TAGTYPE_MATERIALDIFFUSECOLOR;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);				// write tag
	count = sizeof(float) * 4;
	FSWrite(refNum, &count, color);				// write diffuse color
		
		
		/**********************/
		/* WRITE TEXTURE INFO */
		/**********************/
		
	if (srcMat->isTextured)
	{		
		WriteTexture(refNum, srcMat);
	}	
		
}



/************************ WRITE TEXTURE **********************/

static void WriteTexture(short refNum, MaterialType *srcMat)
{
BG3DTextureHeader	textureHeader;
TQ3Status			status;
TQ3TextureObject	texture;
int					w,h,x,y;
TQ3StoragePixmap	pixmap;
TQ3Mipmap			mipmap;
GWorldPtr			textureGWorld;
Ptr					texBuff;
Rect				rect;
TQ3PixelType		pixelType;
PixMapHandle	gworldPixmap;
long			rowBytes,buffSize;
u_short			*srcPtr16;
u_long			*srcPtr32,*destPtr32;
GLint			pixelFormat;
long			count;
u_long			tag;
u_char			*destPtr24;
u_long			r,g,b,a;

				/****************************/
				/* GET INFO ON THIS TEXTURE */
				/****************************/
				
	status = Q3TextureShader_GetTexture(srcMat->textureRef, &texture);								// get texture from shader
	if (status != kQ3Success)
		DoFatalAlert("\pWriteTexture: Q3TextureShader_GetTexture failed!");

	switch(Q3Texture_GetType(texture))														// must be pixmap or mipmap src
	{
		case	kQ3TextureTypePixmap:
				status = Q3PixmapTexture_GetPixmap(texture, &pixmap);						// get texture's pixmap
				if (status != kQ3Success)
					DoFatalAlert("\pWriteTexture: Q3PixmapTexture_GetPixmap failed!");
				textureGWorld = QD3D_PixmapToGWorld(&pixmap,&rect);					
				Q3Object_Dispose(pixmap.image);												// nuke old image storage
				pixelType = pixmap.pixelType;			
				break;
				
		case	kQ3TextureTypeMipmap:
				status = Q3MipmapTexture_GetMipmap(texture, &mipmap);						// get texture's mipmap
				if (status != kQ3Success)
					DoFatalAlert("\pWriteTexture: Q3MipmapTexture_GetMipmap failed!");
				textureGWorld = QD3D_MipmapToGWorld(&mipmap,&rect);							// draw into gworld	
				Q3Object_Dispose(mipmap.image);												// nuke old image storage
				pixelType = mipmap.pixelType;			
				break;
				
		default:					
				DoFatalAlert("\pWarning, WriteTexture enountered a non-pixmap texture.");
	}

	w = rect.right - rect.left;
	h = rect.bottom - rect.top;


		/***********************************/
		/* COPY RAW PIXEL DATA INTO BUFFER */
		/***********************************/

			/* GET GWORLD INFO */
			
	gworldPixmap = GetGWorldPixMap(textureGWorld);
	LockPixels(gworldPixmap);
	rowBytes = (**gworldPixmap).rowBytes & 0x3fff;	// get GWorld's rowbytes
	srcPtr16 = (u_short *)GetPixBaseAddr(gworldPixmap);		// get ptr to pixels
	srcPtr32 = (u_long *)srcPtr16;
	

			/* COPY BASED ON PIXEL TYPE */
			
	switch(pixelType)
	{
		case	kQ3PixelTypeRGB32:
				pixelFormat = GL_RGB;								// set GL pixel format
				texBuff = AllocPtr(buffSize = w * h * 3);			// alloc 24-bit buffer
				if (texBuff == nil)
					DoFatalAlert("\pWriteTexture: AllocPtr failed!");
				destPtr24 = (u_char *)texBuff;
				destPtr24 += (w*3) * (h-1);							// flip texture for OpenGL
		
				for (y = 0; y < h; y++)								// copy pixels to buffer
				{				
					for (x = 0; x < w; x++)
					{
						r = (srcPtr32[x] & 0x00ff0000) >> 16;
						g = (srcPtr32[x] & 0x0000ff00) >> 8;
						b = srcPtr32[x] & 0x000000ff;
						
						destPtr24[x*3]   = r;
						destPtr24[x*3+1] = g;
						destPtr24[x*3+2] = b;
					}
	
					destPtr24 -= w*3;
					srcPtr32 += rowBytes/4;
				}				
				break;
				
		case	kQ3PixelTypeARGB32:
				pixelFormat = GL_RGBA;								// set GL pixel format
				texBuff = AllocPtr(buffSize = w * h * 4);						// alloc buffer
				if (texBuff == nil)
					DoFatalAlert("\pWriteTexture: AllocPtr failed!");
				destPtr32 = (u_long *)texBuff;
				destPtr32 += w * (h-1);								// flip for OpenGL
		
				for (y = 0; y < h; y++)								// copy pixels to buffer
				{				
					for (x = 0; x < w; x++)
					{
						r = (srcPtr32[x] & 0x00ff0000) >> 16;
						g = (srcPtr32[x] & 0x0000ff00) >> 8;
						b = srcPtr32[x] & 0x000000ff;
						a = (srcPtr32[x] & 0xff000000) >> 24;
						
						destPtr32[x] = r << 24;	
						destPtr32[x] |= g << 16;	
						destPtr32[x] |= b << 8;	
						destPtr32[x] |= a;	
					}
					destPtr32 -= w;
					srcPtr32 += rowBytes/4;
				}				
				break;

		case	kQ3PixelTypeRGB16:
				pixelFormat = GL_RGB;								// set GL pixel format
				texBuff = AllocPtr(buffSize = w * h * 3);			// alloc 24 bit buffer
				if (texBuff == nil)
					DoFatalAlert("\pWriteTexture: AllocPtr failed!");
				destPtr24 = (u_char *)texBuff;
				destPtr24 += (w * 3) * (h-1);						// flip for OpenGL
		
				for (y = 0; y < h; y++)								// copy pixels to buffer
				{				
					for (x = 0; x < w; x++)
					{
						u_long	pixel,r,g,b;
						
						pixel = srcPtr16[x];						// get 16-bit pixel
						
						r = (pixel & (0x1f << 10)) >> 7;
						g = (pixel & (0x1f << 5)) >> 2;
						b = (pixel & 0x1f) << 3;
												
						destPtr24[x*3] 	 	= r;					// save 24-bit result			
						destPtr24[(x*3)+1]  = g;
						destPtr24[(x*3)+2]  = b;
					}	
					destPtr24 -= w*3;
					srcPtr16 += rowBytes/2;
				}				
				break;
				
		case	kQ3PixelTypeARGB16:	
				pixelFormat = GL_RGBA;								// set GL pixel format
				texBuff = AllocPtr(buffSize = w * h * 4);			// alloc buffer
				if (texBuff == nil)
					DoFatalAlert("\pWriteTexture: AllocPtr failed!");
				destPtr32 = (u_long *)texBuff;
				destPtr32 += w * (h-1);								// flip for OpenGL
		
				for (y = 0; y < h; y++)								// copy pixels to buffer
				{				
					for (x = 0; x < w; x++)
					{
						u_long	pixel;
						
						pixel = srcPtr16[x];						// get 16-bit pixel
						
						r = (pixel & (0x1f<<10)) >> 7;
						g = (pixel & (0x1f<<5)) >> 2;
						b = (pixel & 0x1f) << 3;
						
						destPtr32[x] = r <<	24;
						destPtr32[x] |= g << 16;
						destPtr32[x] |= b << 8;
						if (pixel & 0x8000)							// see if set alpha
							destPtr32[x] |= 0xff;
					}	
					destPtr32 -= w;
					srcPtr16 += rowBytes/2;
				}				
				break;				
	}



		/* SETUP TEXTURE HEADER */

	textureHeader.width 		= w;
	textureHeader.height 		= h;
	textureHeader.pixelFormat 	= pixelFormat;
	textureHeader.bufferSize 	= buffSize;
	
	textureHeader.reserved[0] = 0;
	textureHeader.reserved[1] = 0;
	textureHeader.reserved[2] = 0;
	textureHeader.reserved[3] = 0;


	/* WRITE TAG, HEADER, AND TEXTURE TO FILE */
	
	tag = BG3D_TAGTYPE_TEXTUREMAP;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);				// write tag
	
	count = sizeof(BG3DTextureHeader);
	FSWrite(refNum, &count, &textureHeader);	// write header

	count = buffSize;
	FSWrite(refNum, &count, texBuff);			// write texture



		/* CLEANUP */
		
	DisposePtr(texBuff);
	DisposeGWorld(textureGWorld);
}


/******************** WRITE BG3D GEOMETRY ***********************/

static void WriteBG3DGeometry(short refNum)
{
TQ3GroupPosition	position;
TQ3Object   		object;
long				count,i;
u_long				tag;

	if (gModelObjNode == nil)
		return;


		/*****************************/
		/* WRITE NEW GROUP START TAG */
		/*****************************/
			
	tag = BG3D_TAGTYPE_GROUPSTART;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);

				
		/****************************************************************/
		/* IF BASE IS GROUP, THEN WRITE OUT OBJECTS IN GROUP SEPARATELY */
		/****************************************************************/
	
	if (Q3Object_IsType(gModelGroup,kQ3ShapeTypeGroup))
 	{
 		i = 0;
  		for (Q3Group_GetFirstPosition(gModelGroup, &position); position != nil; Q3Group_GetNextPosition(gModelGroup, &position))	// scan all objects in group
 		{
   			Q3Group_GetPositionObject (gModelGroup, position, &object);		// get object from group
			if (object != NULL)
   			{
				WriteATriMesh(refNum,i,object);
   				Q3Object_Dispose(object);
   				i++;
   			}
  		}
	}

		/*******************************************************/	
		/* BASE ISNT A GROUP, SO JUST WRITE WHATEVER IT IS OUT */
		/*******************************************************/	
	else
	{
		WriteATriMesh(refNum,0,gModelGroup);
	}

		/***********************/
		/* WRITE END GROUP TAG */
		/***********************/
			
	tag = BG3D_TAGTYPE_GROUPEND;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);

}


/********************** WRITE A TRIMESH **************************/
//
// INPUT: i = material # associated with this trimesh.
//

static void WriteATriMesh(short refNum, int i, TQ3Object triMeshObj)
{
TQ3TriMeshData	triMeshData;
long			count,numVertices,n,numTriangles;
u_long			tag;
BG3DGeometryHeader	header;

	Q3TriMesh_GetData(triMeshObj, &triMeshData);


		/***********************************/
		/* WRITE NEW GEOMETRY TAG & HEADER */
		/***********************************/

	tag = BG3D_TAGTYPE_GEOMETRY;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);

	header.type = BG3D_GEOMETRYTYPE_VERTEXELEMENTS;
	header.numMaterials = 1;							// trimeshes can only have 1 material
	header.materialNum[0] = i;
	header.flags = 0;
	header.numPoints 	= triMeshData.numPoints;
	header.numTriangles = triMeshData.numTriangles;
	header.reserved[0] = 0;
	header.reserved[1] = 0;
	header.reserved[2] = 0;
	header.reserved[3] = 0;

	count = sizeof(BG3DGeometryHeader);
	FSWrite(refNum, &count, &header);


		/**********************/
		/* WRITE VERTEX ARRAY */
		/**********************/
			
			/* WRITE TAG */
			
	tag = BG3D_TAGTYPE_VERTEXARRAY;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);
	

		/* WRITE VERTEX DATA */
		
	numVertices = triMeshData.numPoints;
	count = sizeof(TQ3Point3D) * numVertices;
	FSWrite(refNum, &count, triMeshData.points);


		/**************************/
		/* WRITE ATTRIBUTE ARRAYS */
		/**************************/
		
	for (n = 0; n < triMeshData.numVertexAttributeTypes; n++)
	{
		switch(triMeshData.vertexAttributeTypes[n].attributeType)
		{
			case	kQ3AttributeTypeNormal:
					WriteNormalArray(refNum, triMeshData.vertexAttributeTypes[n].data, numVertices);
					break;
					
			case	kQ3AttributeTypeSurfaceUV:
			case	kQ3AttributeTypeShadingUV:
					WriteUVArray(refNum, triMeshData.vertexAttributeTypes[n].data, numVertices);
					break;
			
			case	kQ3AttributeTypeDiffuseColor:
					WriteColorArray(refNum, triMeshData.vertexAttributeTypes[n].data, numVertices);
					break;
		}
	}
	

		/******************************/
		/* WRITE TRIANGLE INDEX ARRAY */
		/******************************/

			/* WRITE TAG */
			
	tag = BG3D_TAGTYPE_TRIANGLEARRAY;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);


		/* WRITE INDEX DATA */
		
	numTriangles = triMeshData.numTriangles;
	count = sizeof(u_long) * numTriangles * 3;
	FSWrite(refNum, &count, triMeshData.triangles);



			/* CLEANUP */
			
	Q3TriMesh_EmptyData(&triMeshData);
}


/******************* WRITE NORMAL ARRAY **********************/

static void WriteNormalArray(short refNum, TQ3Vector3D *vectors, int numPoints)
{
long			count;
u_long			tag;

			/* WRITE TAG */
			
	tag = BG3D_TAGTYPE_NORMALARRAY;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);
	

		/* WRITE VERTEX DATA */

	count = sizeof(TQ3Vector3D) * numPoints;
	FSWrite(refNum, &count, vectors);
}


/******************* WRITE UV ARRAY **********************/

static void WriteUVArray(short refNum, TQ3Param2D *uvs, int numPoints)
{
long			count;
u_long			tag;

			/* WRITE TAG */
			
	tag = BG3D_TAGTYPE_UVARRAY;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);
	

		/* WRITE VERTEX DATA */

	count = sizeof(TQ3Param2D) * numPoints;
	FSWrite(refNum, &count, uvs);
}


/******************* WRITE COLOR ARRAY **********************/

static void WriteColorArray(short refNum, TQ3ColorRGB *colors, int numPoints)
{
int				i,j;
long			size,count;
u_long			tag;
u_char			*tempBuff;

		/* CONVERT RGB TO RGBA */
		
	size = sizeof(u_char) * numPoints * 4;
	tempBuff = (u_char *)AllocPtr(size);							// alloc buffer to hold rgba BYTES
	if (tempBuff == nil)
		DoFatalAlert("\pWriteColorArray: alloc ptr failed!");
	
	for (i = j = 0; i < numPoints; i++)
	{
		tempBuff[j++] = colors[i].r * 255.0f;
		tempBuff[j++] = colors[i].g * 255.0f;
		tempBuff[j++] = colors[i].b * 255.0f;
		tempBuff[j++] = 255.0;
	}
	

			/* WRITE TAG */
			
	tag = BG3D_TAGTYPE_COLORARRAY;
	count = sizeof(tag);
	FSWrite(refNum, &count, &tag);
	

		/* WRITE VERTEX DATA */

	count = size;
	FSWrite(refNum, &count, tempBuff);
	
	
	DisposePtr((Ptr)tempBuff);
}







