/****************************/
/*   	TRIANGULATOR.C	    */
/* (c)1998 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/****************************/
/*    EXTERNALS             */
/****************************/


#include <QuickDraw.h>
#include <TextUtils.h>
#include <QD3DExtension.h>
#include <QD3DRenderer.h>
#include <QD3DGeometry.h>
#include "myglobals.h"
#include "misc.h"
#include "optimize.h"
#include "triangulator.h"
#include "qd3d_support.h"

extern	WindowPtr		gModelWindow;


/****************************/
/*    PROTOTYPES            */
/****************************/

static TQ3XFunctionPointer TR_MetaHandler(TQ3XMethodType methodType);
static TQ3Status TR_StartFrame(TQ3ViewObject view, void *srPrivate, TQ3DrawContextObject	drawContext);
static TQ3Status TR_Pass(TQ3ViewObject view, void *srPrivate, TQ3CameraObject camera, TQ3GroupObject lightGroup);
static TQ3ViewStatus TR_EndPass(TQ3ViewObject 	view, void *srPrivate);
static TQ3XFunctionPointer TR_Geometry_MetaHandler(TQ3XMethodType	methodType);
static TQ3Status TR_Geometry_Triangle(TQ3ViewObject view, void *srPrivate, TQ3GeometryObject triangle, 
								const TQ3TriangleData	*triangleData);
static void TR_Cancel(TQ3ViewObject view, void *srPrivate);


/****************************/
/*    CONSTANTS             */
/****************************/



/*********************/
/*    VARIABLES      */
/*********************/

QD3DSetupOutputType		gTriangulatorViewInfo;


static TQ3XObjectClass	gRendererClass;
TQ3ObjectType	gClassType;

TQ3SubdivisionStyleData	gSubdivData;

#pragma mark ======== RENDERER SETUP & REGISTER ===========


/***************** CREATE MY TRIANGULATOR RENDERER ***********************/
//
// I create a plug-in renderer which will receive the decomposed QD3D Geometries
// as TriMeshes and Triangles.
//

void CreateMyTriangulatorRenderer(void)
{
QD3DSetupInputType	viewDef;
TQ3Point3D			cameraFrom = { 0,8, 00 };
TQ3Point3D			cameraTo = { 0.0, 0, 0 };
TQ3Vector3D			cameraUp = { 0.0, 0.0, -1.0 };
TQ3ColorARGB		clearColor = {1,0,0,.5};
TQ3ColorRGB			ambientColor = { 1.0, 1.0, 1.0 };

				/* CREATE THE NEW PLUGIN RENDERER CLASS */
				
	gRendererClass = Q3XObjectHierarchy_RegisterClass(kQ3SharedTypeRenderer,
														&gClassType, "TriangulatorRenderer",
														TR_MetaHandler,	NULL, 0,
														0);
	if (gRendererClass == nil)
	{
		DoAlert("\pCreateMyTriangulatorRenderer: Q3XObjectHierarchy_RegisterClass failed!");
		QD3D_ShowRecentError();
	}
	
			/* NOW LETS CREATE A NEW VIEW TO USE FOR RENDERING WITH THIS RENDERER */
			
	viewDef.view.displayWindow 		= gModelWindow;
	viewDef.view.rendererType 		= gClassType;
	viewDef.view.clearColor 		= clearColor;
	viewDef.view.paneClip.left 		= 0;
	viewDef.view.paneClip.right 	= 0;
	viewDef.view.paneClip.top 		= 0;
	viewDef.view.paneClip.bottom 	= 0; 

	viewDef.styles.interpolation 	= kQ3InterpolationStyleVertex; 
	viewDef.styles.backfacing 		= kQ3BackfacingStyleRemove;
	viewDef.styles.fill				= kQ3FillStyleFilled; 
	viewDef.styles.shaderType 		= kQ3IlluminationTypeNULL; 

	viewDef.camera.from 			= cameraFrom;
	viewDef.camera.to 				= cameraTo;
	viewDef.camera.up 				= cameraUp;
	viewDef.camera.hither 			= 0.5;
	viewDef.camera.yon 				= 500;
	viewDef.camera.fov 				= 1.0;

	viewDef.lights.ambientBrightness = 0.1;
	viewDef.lights.ambientColor 	= ambientColor;
	viewDef.lights.numFillLights 	= 0;

	QD3D_SetupWindow(&viewDef, &gTriangulatorViewInfo);
	
	
			/* SET TESSELATION / SUBDIVISION STYLE */
			
	gSubdivData.method = kQ3SubdivisionMethodConstant;
	gSubdivData.c1 = gSubdivData.c2 = DEFAULT_TESS;
}



/********************** TR: METAHANDLER **************************/
//
//	Main metahandler. Provides entry point for various
//	required and optional renderer functions, and for the
//	metahandlers for geometric primitive rendering and
//	style, matrix, and attribute updates.
//

static TQ3XFunctionPointer TR_MetaHandler(TQ3XMethodType methodType)
{
	switch (methodType)
	{				
		 		/* RENDERER */
		 
		case 	kQ3XMethodTypeRendererStartFrame:
				return (TQ3XFunctionPointer) TR_StartFrame;
		
		case 	kQ3XMethodTypeRendererStartPass:
				return (TQ3XFunctionPointer) TR_Pass;
	
		case 	kQ3XMethodTypeRendererEndPass:
				return (TQ3XFunctionPointer) TR_EndPass;
					
		case 	kQ3XMethodTypeRendererCancel:
				return (TQ3XFunctionPointer) TR_Cancel;
					

			 	/* RENDERER DRAW */
		 
		case 	kQ3XMethodTypeRendererSubmitGeometryMetaHandler:
				return (TQ3XFunctionPointer) TR_Geometry_MetaHandler;
	

		 		/* "IS INTERACTIVE" CONTROL */
		 
		case 	kQ3XMethodTypeRendererIsInteractive:
				return (TQ3XFunctionPointer) kQ3True;
			
		default:
				return NULL;
	}
}


/********************* TR: START FRAME **************************/

static TQ3Status TR_StartFrame(TQ3ViewObject view, void *srPrivate, TQ3DrawContextObject	drawContext)
{
	return (kQ3Success);
}


/********************* TR: PASS **************************/

static TQ3Status TR_Pass(TQ3ViewObject view, void *srPrivate, TQ3CameraObject camera, TQ3GroupObject lightGroup)
{
	return (kQ3Success);
}


/***************** TR: END  PASS ***********************/

static TQ3ViewStatus TR_EndPass(TQ3ViewObject 	view, void *srPrivate)
{
	return (kQ3ViewStatusDone);
}


/**************** TR: CANCEL ********************/

static void TR_Cancel(TQ3ViewObject view, void *srPrivate)
{
}



/********************* TR: GEOMETRY METAHANDLER ***********************/
//
//	Provide entry points for geometric primitive rendering functions.
//	We only want to support Triangles.  Returning NULL for all others
//  will trick QD3D into decomposing more complex geometries into triangles.
//

static TQ3XFunctionPointer TR_Geometry_MetaHandler(TQ3XMethodType	methodType)
{
	switch (methodType)
	{	
		case 	kQ3GeometryTypeTriangle:
				return (TQ3XFunctionPointer) TR_Geometry_Triangle;

		default:
				return NULL;
	}
}


#pragma mark ========= DECOMPOSED DATA ENTRY POINTS ===========


/******************** TR: GEOMETRY TRIANGLE **************************/
//
// We simply take the Triangle object and pass it to the regular old triangle optimizing stuff.
//
// NOTE: I found that if triangle == nil and triangleData has stuff and then I transform the
//		points in triangleData, then weird and bad things happen.  So, to avoid this,
//		I take the triangleData and convert it into a triangleObject, thus duplicating the data
//		so that the transform function can mess with it all it wants.
//

static TQ3Status TR_Geometry_Triangle(TQ3ViewObject view, void *srPrivate, TQ3GeometryObject triangle, 
								const TQ3TriangleData	*triangleData)
{		
	if (triangle == nil)
	{
		triangle = Q3Triangle_New(triangleData);
		PreTransformTri(triangle,nil);
		AddTriToTriangleLists(triangle,nil);
		Q3Object_Dispose(triangle);
	}
	else
	{
		PreTransformTri(triangle,triangleData);
		AddTriToTriangleLists(triangle,triangleData);
	}
	return (kQ3Success);
}



#pragma mark ============= DECOMPOSE FUNCTION =====================

/********************** DECOMPOSE GEOMETRY ****************************/
//
// By submitting the geometry to our triangulator renderer, we can decompose
// the thing down into triangles or TriMeshes which our application knows how to deal with.
//

void DecomposeGeometry(TQ3Object geometry)
{
TQ3Status				myStatus;
TQ3ViewStatus			myViewStatus;

			/* START RENDERING */

			
	myStatus = Q3View_StartRendering(gTriangulatorViewInfo.viewObject);			
	if ( myStatus == kQ3Failure )
		DoFatalAlert("\pDecomposeGeometry: Q3View_StartRendering Failed!");

	do
	{
		Q3SubdivisionStyle_Submit(&gSubdivData, gTriangulatorViewInfo.viewObject);
		
		myStatus = Q3Object_Submit(geometry,gTriangulatorViewInfo.viewObject);
		if ( myStatus == kQ3Failure )
			DoFatalAlert("\p Q3Style_Submit Failed!");

		myViewStatus = Q3View_EndRendering(gTriangulatorViewInfo.viewObject);
		
	} while ( myViewStatus == kQ3ViewStatusRetraverse );
}












